// Copyright (c) 2011, Giliam de Carpentier. All rights reserved.
// See http://www.decarpentier.nl/scape-procedural-basics/ for more info
// 
// This file is licensed under the FreeBSD license:
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
//    1. Redistributions of source code must retain the above copyright notice, this list of
//       conditions and the following disclaimer.
// 
//    2. Redistributions in binary form must reproduce the above copyright notice, this list
//       of conditions and the following disclaimer in the documentation and/or other materials
//       provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GILIAM DE CARPENTIER 'AS IS' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GILIAM DE CARPENTIER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "ProceduralBasicsSAS.h"

#define OCTAVES 10 // using a 'static' value here to prevent the need for dynamic loop branching

float perlinNoise(float2 p, float seed)
{
	// Calculate 2D integer coordinates i and fraction p.
	float2 i = floor(p);
	float2 f = p - i;

	// Get weights from the coordinate fraction
	float2 w = f * f * f * (f * (f * 6 - 15) + 10);
	float4 w4 = float4(1, w.x, w.y, w.x * w.y);

	// Get the four randomly permutated indices from the noise lattice nearest to
	// p and offset these numbers with the seed number.
	float4 perm = tex2D(samplerPerlinPerm2D, i / 256) + seed;

	// Permutate the four offseted indices again and get the 2D gradient for each
	// of the four permutated coordinates-seed pairs.
	float4 g1 = tex2D(samplerPerlinGrad2D, perm.xy) * 2 - 1;
	float4 g2 = tex2D(samplerPerlinGrad2D, perm.zw) * 2 - 1;

	// Evaluate the four lattice gradients at p
	float a = dot(g1.xy, f);
	float b = dot(g2.xy, f + float2(-1,  0));
	float c = dot(g1.zw, f + float2( 0, -1));
	float d = dot(g2.zw, f + float2(-1, -1));

	// Bi-linearly blend between the gradients, using w4 as blend factors.
	float4 grads = float4(a, b - a, c - a, a - b - c + d);
	float n = dot(grads, w4);

	// Return the noise value, roughly normalized in the range [-1, 1]
	return n * 1.5;
}

float turbulence(float2 p, float seed, int octaves, float lacunarity = 2.0, float gain = 0.5)
{
	float sum = 0;
	float freq = 1.0, amp = 1.0;
	for (int i=0; i < octaves; i++)
	{
		float n = perlinNoise(p*freq, seed + i / 256.0);
		sum += n*amp;
		freq *= lacunarity;
		amp *= gain;
	}
	return sum;
}

float billowedNoise(float2 p, float seed)
{
	return abs(perlinNoise(p, seed));
}

float billowyTurbulence(float2 p, float seed, int octaves, float lacunarity = 2.0, float gain = 0.5)
{
	float sum = 0;
	float freq = 1.0, amp = 1.0;
	for (int i=0; i < octaves; i++)
	{
		float n = billowedNoise(p*freq, seed + i / 256.0);
		sum += n*amp;
		freq *= lacunarity;
		amp *= gain;
	}
	return sum;
}

float ridgedNoise(float2 p, float seed)
{
	return 1.0f-abs(perlinNoise(p, seed));
}

float ridgedTurbulence(float2 p, float seed, int octaves, float lacunarity = 2.0, float gain = 0.5)
{
	float sum = 0;
	float freq = 1.0, amp = 1.0;
	for (int i=0; i < octaves; i++)
	{
		float n = ridgedNoise(p*freq, seed + i / 256.0);
		sum += n*amp;
		freq *= lacunarity;
		amp *= gain;
	}
	return sum;
}

float3 perlinNoisePseudoDeriv(float2 p, float seed)
{
	// Calculate 2D integer coordinates i and fraction p.
	float2 i = floor(p);
	float2 f = p - i;
	
	// Get weights from the coordinate fraction
	float2 w = f * f * f * (f * (f * 6 - 15) + 10);
	float4 w4 = float4(1, w.x, w.y, w.x * w.y);

	// Get pseudo derivative weights
	float2 dw = f * f * (f * (30 * f - 60) + 30);
	
	// Get the four randomly permutated indices from the noise lattice nearest to
	// p and offset these numbers with the seed number.
	float4 perm = tex2D(samplerPerlinPerm2D, i / 256) + seed;
	
	// Permutate the four offseted indices again and get the 2D gradient for each
	// of the four permutated coordinates-seed pairs.
	float4 g1 = tex2D(samplerPerlinGrad2D, perm.xy) * 2 - 1;
	float4 g2 = tex2D(samplerPerlinGrad2D, perm.zw) * 2 - 1;
	
	// Evaluate the four lattice gradients at p
	float a = dot(g1.xy, f);
	float b = dot(g2.xy, f + float2(-1,  0));
	float c = dot(g1.zw, f + float2( 0, -1));
	float d = dot(g2.zw, f + float2(-1, -1));
	
	// Bi-linearly blend between the gradients, using w4 as blend factors.
	float4 grads = float4(a, b - a, c - a, a - b - c + d);
	float n = dot(grads, w4);

	// Calculate pseudo derivates
	float dx = dw.x * (grads.y + grads.w*w.y);
	float dy = dw.y * (grads.z + grads.w*w.x);

	// Return the noise value, roughly normalized in the range [-1, 1]
	// Also return the pseudo dn/dx and dn/dy, scaled by the same factor
	return float3(n, dx, dy) * 1.5;
}

float iqTurbulence(float2 p, float seed, int octaves, float lacunarity = 2.0, float gain = 0.5)
{
	float sum = 0.5;
	float freq = 1.0, amp = 1.0;
	float2 dsum = float2(0,0);
	for (int i=0; i < octaves; i++)
	{
		float3 n = perlinNoisePseudoDeriv(p*freq, seed + i / 256.0);
		dsum += n.yz;
		sum += amp * n.x / (1 + dot(dsum, dsum));
		freq *= lacunarity;
		amp *= gain;
	}
	return sum;
}

struct vertexInput {
    float3 Position	: POSITION;
    float4 UV		: TEXCOORD0;
};

struct vertexOutput {
    float4 HPosition: POSITION;
    float2 UV		: TEXCOORD0;
};

vertexOutput vsSimple(vertexInput IN)
{
    float4 Po = float4(IN.Position.xyz, 1);
	vertexOutput OUT;
#if DIRECT3D_VERSION
    OUT.HPosition = mul(Po, worldViewProjection);
#else // DIRECT3D_VERSION
    OUT.HPosition = mul(worldViewProjection, Po);
#endif // DIRECT3D_VERSION
    OUT.UV = IN.UV.xy;
    return OUT;
}

float4 psPerlinNoise(vertexOutput IN) : COLOR
{
	float height = perlinNoise(IN.UV * gUVScale, gSeed);
	return height * 0.5f + 0.5f;
}

float4 psTurbulence(vertexOutput IN) : COLOR
{
	float height = turbulence(IN.UV * gUVScale, gSeed, OCTAVES, gLacunarity, gGain);
	return height * 0.4f + 0.5f;
}

float4 psBillowyTurbulence(vertexOutput IN) : COLOR
{
	float height = billowyTurbulence(IN.UV * gUVScale, gSeed, OCTAVES, gLacunarity, gGain);
	return height * 0.5f + 0.25f;
}

float4 psRidgedTurbulence(vertexOutput IN) : COLOR
{
	float height = ridgedTurbulence(IN.UV * gUVScale, gSeed, OCTAVES, gLacunarity, gGain);
	return height * 0.5f - 0.33f;
}

float4 psIqTurbulence(vertexOutput IN) : COLOR
{
	float height = iqTurbulence(IN.UV * gUVScale, gSeed, OCTAVES, gLacunarity, gGain);
	return height * 0.4f + 0.33f;
}
